import io
import logging
from typing import TYPE_CHECKING, Dict, Iterator, List, Optional, Union

import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc

from odp.tabular_v2 import big, bsquare
from odp.tabular_v2.client import Client
from odp.tabular_v2.util import exp, vars_to_json

if TYPE_CHECKING:
    from odp.tabular_v2.client import Cursor


class Table:
    def __init__(self, client: Client, table_id: str):
        self._id = table_id
        self._client = client
        self._inner_schema: Optional[pa.Schema] = None
        self._outer_schema: Optional[pa.Schema] = None
        self._tx = None
        try:
            self._fetch_schema()
        except FileNotFoundError:
            pass  # ktable does not exist yet

        self._bigcol = big.RemoteBigCol(
            self._bigcol_upload,
            self._bigcol_download,
            "/tmp/cache_big/",  # FIXME
        )

    def _fetch_schema(self):
        if self._inner_schema:
            return
        empty = list(self._select(inner_query=exp.parse('"fetch" == "schema"')))
        assert len(empty) == 1
        assert empty[0].num_rows == 0
        self._inner_schema = empty[0].schema
        mid = big.convert_schema_outward(self._inner_schema)
        self._outer_schema = bsquare.convert_schema_outward(mid)

    def drop(self):
        """
        drop the table data and schema
        this operation is irreversible
        @return:
        """
        try:
            res = self._client._request(
                path="/api/table/v2/drop",
                params={
                    "table_id": self._id,
                },
            ).json()
            logging.info("dropped %s: %s", self._id, res)
        except FileNotFoundError:
            logging.info("table %s does not exist", self._id)

    def _validate_parquet_schema(self, schema: pa.Schema):
        compatible_types = {
            pa.types.is_integer,
            pa.types.is_floating,
            pa.types.is_boolean,
            pa.types.is_string,
            pa.types.is_binary,
            pa.types.is_date,
            pa.types.is_timestamp,
            pa.types.is_decimal,
            pa.types.is_time,
        }
        for field in schema:
            if not any(check_function(field.type) for check_function in compatible_types):
                raise ValueError(f"Incompatible type for parquet detected: {field.name} ({field.type})")

    def create(self, schema: pa.Schema):
        """
        set the table schema using the given pyarrow schema
        fields might contains metadata which will be used internally:
        * index: the field should be used to partition the data
        * isGeometry: the field is a geometry (wkt for string, wkb for binary)
        @param schema: pyarrow.Schema
        @raise FileExistsError if the schema is already set
        @return:
        """
        self._validate_parquet_schema(schema)
        self._outer_schema = schema
        schema = bsquare.convert_schema_inward(schema)
        self._inner_schema = big.convert_schema_inward(schema)
        buf = io.BytesIO()
        w = pa.ipc.RecordBatchStreamWriter(buf, self._inner_schema)
        w.write_batch(pa.RecordBatch.from_pylist([], schema=self._inner_schema))
        w.close()

        self._client._request(
            path="/api/table/v2/create",
            params={
                "table_id": self._id,
            },
            data=buf.getvalue(),
        ).json()

    def _bigcol_upload(self, bid: str, data: bytes):
        self._client._request(
            path="/api/table/v2/big_upload",
            params={"table_id": self._id, "big_id": bid},
            data=data,
        )

    def _bigcol_download(self, big_id: str) -> bytes:
        return self._client._request(
            path="/api/table/v2/big_download",
            params={"table_id": self._id, "big_id": big_id},
        ).all()

    # def aggregate_h3(
    #    self, geometry: shapely.Geometry, resolution: int = 0, query: Union[exp.Op, str, None] = None
    # ) -> dict:
    #    """
    #    experimental
    #    aggregate data using h3 hexagons withing the given geometry
    #    compute the minimal resolution to have at least 200 hexagons
    #    """
    #    import h3

    #    pol = json.loads(shapely.to_geojson(geometry))
    #    hexes = h3.polyfill(pol, resolution)
    #    if resolution == 0:
    #        for resolution in range(1, 15):
    #            hexes = h3.polyfill(pol, resolution)
    #            if len(hexes) >= 20:
    #                logging.info("using res %d (will give %d hexagons)", resolution, len(hexes))
    #                break

    #    if len(hexes) > 10_0000:
    #        raise ValueError(f"too many hexagons: {len(hexes)}, please use a lower resolution")

    #    hpol = h3.h3_set_to_multi_polygon(hexes, geo_json=True)[0]
    #    pols = shapely.polygons(hpol)
    #    bounds = pols[0].bounds

    #    geo_query = exp.parse(f"lat > {bounds[1]} and lat < {bounds[3]} and lon > {bounds[0]} and lon < {bounds[2]}")
    #    logging.info("geo_query: %s", geo_query)
    #    if query:
    #        if isinstance(query, str):
    #            query = exp.parse(query)
    #        query = exp.BinOp(
    #            left=query,  # TODO: do we need to wrap it in a parenthesis?
    #            op="and",
    #            right=geo_query,
    #        )
    #    else:
    #        query = geo_query

    #    out = {}
    #    for row in self.aggregate(by="h3(%d)" % resolution, query=query):
    #        hid = row["__aggr"]
    #        del row["__aggr"]
    #        if hid in hexes:
    #            out[hid] = row
    #    return out

    def aggregate(
        self,
        by: str,
        query: Union[exp.Op, str, None] = None,
        aggr: Union[dict, None] = None,
        timeout: float = 30.0,
        vars: Union[dict, list, None] = None,
    ) -> pd.DataFrame:
        """
        aggregate the data after the optional `query` filter
        the paramater `by` is used to determine the key for the aggregation, and can be an expression.
        the optional `aggr` specify which fields need to be aggregated, and how
        If not specified, the fields with metadata "aggr" will be used
        a single DataFrame will be returned, with the index set to the key used for aggregation
        """
        if isinstance(query, str):
            query = exp.parse(query)  # NOTE(oha): we don't do the bsquare and bigcol waltz, but neither the server :(

        if aggr is None:
            aggr = {}
            for field in self._outer_schema:
                if field.metadata and b"aggr" in field.metadata:
                    aggr[field.name] = field.metadata[b"aggr"].decode()

        tot_func = {}
        for field, a_type in aggr.items():
            if a_type == "mean" or a_type == "avg":
                tot_func[field + "_sum"] = "sum"
                tot_func[field + "_count"] = "sum"
            elif a_type == "sum":
                tot_func[field + "_sum"] = "sum"
            elif a_type == "min":
                tot_func[field + "_min"] = "min"
            elif a_type == "max":
                tot_func[field + "_max"] = "max"
            elif a_type == "count":
                tot_func[field + "_count"] = "sum"
            else:
                raise ValueError(f"unknown aggregation type: {a_type}")

        total: Union[pd.DataFrame, None] = None
        for b in self._select(type="aggregate", by=by, inner_query=query, timeout=timeout, aggr=aggr, vars=vars):
            df: pd.DataFrame = b.to_pandas()
            # logging.warning("PARTIAL:\n%s", df)
            if total is None:
                total = df
            else:
                total = pd.concat([total, df], ignore_index=True)
                total = total.groupby("").agg(tot_func).reset_index()
        if total is None:
            return pd.DataFrame()

        for field, a_type in aggr.items():
            logging.info("field: %s, type: %s", field, a_type)
            if a_type == "mean" or a_type == "avg":
                total[field] = total[field + "_sum"] / total[field + "_count"]
                total.drop(columns=[field + "_sum", field + "_count"], inplace=True)
            elif a_type in "sum":
                total[field] = total[field + "_sum"]
                total.drop(columns=[field + "_sum"], inplace=True)
            elif a_type == "min":
                total[field] = total[field + "_min"]
                total.drop(columns=[field + "_min"], inplace=True)
            elif a_type == "max":
                total[field] = total[field + "_max"]
                total.drop(columns=[field + "_max"], inplace=True)
            elif a_type == "count":
                total[field] = total[field + "_count"]
                total.drop(columns=[field + "_count"], inplace=True)
            else:
                raise ValueError(f"unknown aggregation type: {a_type}")

        total = total.set_index("")
        # logging.info("TOTAL:\n%s", total)
        return total

    def _query_cursor(
        self,
        query: Union[exp.Op, str, None] = None,
        cols: Optional[List[str]] = None,
        vars: Union[dict, list, None] = None,
        stream_ttl: float = 30.0,
        type: str = "select",
        tx_id: str = "",
    ) -> "Cursor":
        if isinstance(query, str):
            query = exp.parse(query)
        elif isinstance(query, pc.Expression):
            query = exp.parse(str(query))
        if query:
            query.bind(vars)
            if cols:
                logging.info("cols: %s", cols)
                inner_cols = set(cols)  # add filtering columns
                for op in query.all():
                    if isinstance(op, exp.Field):
                        inner_cols.add(str(op))
                inner_cols = list(inner_cols)
            else:
                inner_cols = None
            query.bind(vars)

            logging.info("outer query: %s", query)
            inner_query = bsquare.convert_query(self._outer_schema, query)
            logging.info("bsquare query: %s", inner_query)
            inner_query = big.inner_exp(self._inner_schema, inner_query)
            logging.info("bigcol query: %s", inner_query)
        else:
            inner_cols = list(cols) if cols else None
            inner_query = None

        # expand the list to include the refs, if not empty
        if inner_cols:
            refs = []
            for col in inner_cols:
                f = self._outer_schema.field(col)
                if f.metadata and b"big" in f.metadata:
                    refs.append(col + ".ref")
            logging.info("inner_cols: %s, refs: %s", inner_cols, refs)
            inner_cols.extend(refs)

        def scanner(scanner_cursor: str) -> Iterator[pd.DataFrame]:
            logging.info("selecting with cursor %s: %s", scanner_cursor, inner_query)
            for b in self._select(
                tx=tx_id,
                type=type,
                inner_query=inner_query,
                cols=inner_cols,
                vars=vars,
                cursor=scanner_cursor,
                timeout=stream_ttl,
            ):
                # logging.info("got %d rows, decoding...", b.num_rows)

                # repackage into small batches because bigcol can use lots of memory
                tab = pa.Table.from_batches([b], schema=b.schema)
                del b
                for b in tab.to_batches(max_chunksize=2_000):
                    df = b.to_pandas()
                    df = self._bigcol.decode(df)
                    df = bsquare.decode(df)
                    if query:
                        df = df[query.pandas(df, self._outer_schema)]
                    if cols:
                        df = df[cols]
                    yield df

        from odp.tabular_v2.client import Cursor

        schema = self._outer_schema
        if cols:
            schema = pa.schema([schema.field(col) for col in cols])
        return Cursor(scanner=scanner, schema=schema)

    def select(
        self,
        query: Union[exp.Op, str, None] = None,
        cols: Optional[List[str]] = None,
        vars: Union[dict, list, None] = None,
        timeout: float = 30.0,
    ) -> "Cursor":
        """
        fetch data from the underling table

        for row in tab.select("age > 18").rows():
            print(row)

        you can use bind variables, especially if you need to use date/time objects:

        for row in tab.select("age > $age", vars={"age": 18}).rows():
            print(row)

        and limits which columns you want to retrieve:

        for row in tab.select("age > 18", cols=["name", "age"]).rows():
            print(row)

        The object returned is a cursor, which can be scanned by rows, batches, pages, pandas dataframes, etc.

        you can check the documentation of the Cursor for more information
        """
        return self._query_cursor(query, cols, vars, timeout, "select")

    def __enter__(self):
        if self._inner_schema is None:
            raise FileNotFoundError("table does not exist, create() it first")

        if self._tx:
            raise ValueError("already in a transaction")

        res = self._client._request(
            path="/api/table/v2/begin",
            params={
                "table_id": self._id,
            },
        ).json()
        from odp.tabular_v2.client.table_tx import Transaction

        buf = big.Buffer(self._bigcol).with_inner_schema(self._inner_schema)
        self._tx = Transaction(self, res["tx_id"], buf)
        return self._tx

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            # logging.warning("aborting transaction %s", self._tx._id)
            # try:
            #    self._client._request(
            #        path="/api/table/v2/rollback",
            #        params={
            #            "table_id": self._id,
            #            "tx_id": self._tx._id,
            #        },
            #    )
            # except Exception as e:
            #    logging.error("ignored: rollback failed: %s", e)
            return
        self._tx.flush()
        self._tx._big_buf.flush()  # flush the bigcol buffer
        self._client._request(
            path="/api/table/v2/commit",
            params={
                "table_id": self._id,
                "tx_id": self._tx._id,
            },
        )
        self._tx = None

    def schema(self) -> pa.Schema:
        """
        return the schema of the table, or None if the schema is not set
        """
        return self._outer_schema

    # used as a filter in Cursor, encode in tx
    def _decode(self, b: pa.RecordBatch) -> pa.RecordBatch:
        b = self._bigcol.decode(b)  # convert to panda first, then do the magic
        b = bsquare.decode(b)
        return b

    def _select(
        self,
        tx: str = "",
        type: str = "select",
        inner_query: Optional[exp.Op] = None,
        aggr: Optional[dict] = None,
        cols: Optional[List[str]] = None,
        vars: Union[Dict, List, None] = None,
        by: Optional[str] = None,
        cursor: str = "",
        timeout: float = 30.0,
    ) -> Iterator[pa.RecordBatch]:
        # t0 = time.perf_counter()
        initialized = False
        while not initialized or cursor:
            # do while - run at least once
            initialized = True
            res = self._client._request(
                path="/api/table/v2/" + type,
                params={
                    "table_id": self._id,
                    "tx_id": tx,
                },
                data={
                    "query": str(inner_query) if inner_query else None,
                    "cols": cols,
                    "cursor": cursor,
                    "aggr": aggr,
                    "by": by,
                    "vars": vars_to_json(vars),
                    "timeout": timeout,
                },
            )
            reader = res.reader()
            r = pa.ipc.RecordBatchStreamReader(reader)
            for bm in r.iter_batches_with_custom_metadata():
                if bm.custom_metadata and b"error" in bm.custom_metadata:
                    raise Exception("server error: %s" % bm.custom_metadata[b"error"].decode())

                if bm.custom_metadata and b"cursor" in bm.custom_metadata:
                    cursor = bm.custom_metadata[b"cursor"].decode()
                    logging.warning("response is partially processed with cursor %s", cursor)
                else:
                    # logging.info("no cursor received on table client, setting to empty")
                    cursor = None

                # logging.info("got batch with %d rows", bm.batch.num_rows)
                yield bm.batch

    def _insert_batch(
        self,
        data: pa.RecordBatch,
        tx: str = "",
    ):
        assert self._inner_schema
        buf = io.BytesIO()
        w = pa.ipc.RecordBatchStreamWriter(buf, self._inner_schema)
        w.write_batch(data)
        w.close()

        self._client._request(
            path="/api/table/v2/insert",
            params={
                "table_id": self._id,
                "tx_id": tx,
            },
            data=buf.getvalue(),
        ).json()



from typing import Optional

import pandas as pd
import pyarrow as pa
import shapely
from shapely import GEOSException

from odp.tabular_v2.bsquare.query import _QueryContext
from odp.tabular_v2.util import exp


def convert_schema_inward(outer_schema: pa.Schema) -> pa.Schema:
    out = []
    for name in outer_schema.names:
        f = outer_schema.field(name)
        if f.metadata and b"isGeometry" in f.metadata:
            meta = f.metadata
            if b"index" in meta:
                new_meta = meta.copy()
                del new_meta[b"index"]
                f = f.with_metadata(new_meta)
            out.append(f)
            out.append(pa.field(name + ".x", pa.float64(), True, metadata=meta))
            out.append(pa.field(name + ".y", pa.float64(), True, metadata=meta))
            out.append(pa.field(name + ".q", pa.float64(), True, metadata=meta))
        else:
            out.append(f)
    return pa.schema(out)


# convert the inner_schema to outer_schema
def convert_schema_outward(inner_schema: pa.Schema) -> pa.Schema:
    geo_indexes = set()

    def is_subfield(schema: pa.Schema, f: pa.Field) -> bool:
        if "." not in f.name:
            return False
        left, right = f.name.rsplit(".", 1)
        if right not in ["x", "y", "q"]:
            return False
        if left not in schema.names:
            return False
        if schema.field(left).metadata and b"isGeometry" not in schema.field(left).metadata:
            return False
        if f.metadata and b"index" in f.metadata:
            geo_indexes.add(left)
        return True

    # create a new schema with only the fields that are not subfields
    fields = []
    for names in inner_schema.names:
        f = inner_schema.field(names)
        if not is_subfield(inner_schema, f):
            fields.append(f)

    # add back the "index" to the main field (which was removed when creating the subfields)
    for i, f in enumerate(fields):
        if f.name in geo_indexes:
            meta = f.metadata
            meta[b"index"] = b"1"
            fields[i] = f.with_metadata(meta)
    return pa.schema(fields)


# convert outer query to inner query using bsquare in .x, .y and .q
def convert_query(outer_schema: pa.Schema, outer_query: Optional[exp.Op]) -> Optional[exp.Op]:
    if outer_query is None:
        return None

    geo_fields = []
    for f in outer_schema:
        if f.metadata and b"isGeometry" in f.metadata:
            geo_fields.append(f.name)

    return _QueryContext(geo_fields).convert(outer_query)


def decode(df: pd.DataFrame) -> pd.DataFrame:
    df = df.copy()
    for name in df.columns:
        if name.endswith(".x") or name.endswith(".y") or name.endswith(".q"):
            df.drop(columns=[name], inplace=True)
    return df


def encode(df: pd.DataFrame, outer_schema: pa.Schema) -> pd.DataFrame:
    # logging.info("bsquare encoding %d rows", b.num_rows)
    geo_names = []
    for name in outer_schema.names:
        f = outer_schema.field(name)
        if f.metadata and b"isGeometry" in f.metadata:
            geo_names.append(name)

    df = df.copy()

    # we encode rows by rows to made it simple to create multiple columns
    def _encode(row):
        for name in geo_names:
            if name in row and row[name] is not None:
                val = row[name]
                if isinstance(val, str):
                    try:
                        val = shapely.from_wkt(val)
                    except GEOSException as e:
                        raise ValueError(f"Invalid geo format: {val}") from e
                elif isinstance(val, bytes):
                    val = shapely.from_wkb(val)
                else:
                    raise ValueError(f"Unsupported type: {type(val)}")
                min_x, min_y, max_x, max_y = val.bounds
                row[name + ".x"] = (min_x + max_x) / 2
                row[name + ".y"] = (min_y + max_y) / 2
                row[name + ".q"] = max(max_x - min_x, max_y - min_y) / 2
            else:
                row[name + ".x"] = None
                row[name + ".y"] = None
                row[name + ".q"] = None
        return row

    for geo_name in geo_names:
        df[geo_name + ".x"] = None
        df[geo_name + ".y"] = None
        df[geo_name + ".q"] = None
    return df.apply(func=_encode, axis=1)


class BSquare:
    geometry_fields = ["{col_name}.x", "{col_name}.y", "{col_name}.q"]  # add complexity and confuse the user

    def __init__(self, inner_schema: Optional[pa.Schema] = None):
        assert not "good"
        self._inner_schema = inner_schema
        self._geo_fields = []  # FIXME do this earlier, then cash on it

    # def _select_filter(self, outer_query: Op | None, schema: pa.Schema | None, cursor: str)
    # -> Iterator[pa.RecordBatch]:
    #    if outer_query is None or outer_query == Scalar.from_py(None):  # FIXME(oha): Scalar(null) is not None
    #        for b in self.inner.select(None, schema, cursor):
    #            yield b
    #        return

    #    geo_fields = []
    #    if schema is None:
    #        schema = BSquare.external_schema(self.inner_schema)
    #    for f in schema:
    #        if f.metadata and b"isGeometry" in f.metadata:
    #            geo_fields.append(f.name)

    #    inner_query = _QueryContext(geo_fields).convert(outer_query)

    #    for b in self.inner.select(
    #        inner_query, self.inner_schema, cursor
    #    ):  # FIXME(oha): expand schema with geo subfields or bigcol will fail
    #        if b.num_rows == 0:
    #            yield b
    #        else:
    #            for b in pa.Table.from_batches([b], schema=b.schema).filter(outer_query.pyarrow()).to_batches():
    #                yield b

    # @staticmethod
    # def external_schema(inner_schema: pa.Schema) -> pa.Schema:
    #    geo_indexes = set()

    #    def is_subfield(schema: pa.Schema, f: pa.Field) -> bool:
    #        if "." not in f.name:
    #            return False
    #        left, right = f.name.rsplit(".", 2)
    #        if left not in schema.names:
    #            return False
    #        if schema.field(left).metadata and b"isGeometry" not in schema.field(left).metadata:
    #            return False
    #        if f.metadata and b"index" in f.metadata:
    #            geo_indexes.add(left)
    #        return True

    #    fields = []
    #    for names in inner_schema.names:
    #        f = inner_schema.field(names)
    #        if not is_subfield(inner_schema, f):
    #            fields.append(f)

    #    for i, f in enumerate(fields):
    #        if f.name in geo_indexes:
    #            meta = f.metadata
    #            meta[b"index"] = b"1"
    #            fields[i] = f.with_metadata(meta)
    #    return pa.schema(fields)

    # def select(self, query: Op | None, schema: pa.Schema | None | List[str], cursor: str) -> Iterator[pa.RecordBatch]:
    #    if schema is not None:
    #        if isinstance(schema, list):
    #            s = self.inner_schema
    #            schema = pa.schema([s.field(n) for n in schema])
    #        for x in self._select_filter(query, schema, cursor):
    #            yield x
    #        return

    #    for b in self._select_filter(query, None, cursor):
    #        if schema is None:
    #            self._inner_schema = b.schema
    #            schema = BSquare.external_schema(self._inner_schema)
    #        b = b.select(schema.names)
    #        yield b

    # def insert(self, batch: pa.RecordBatch):
    #    # NOTE(oha): we trust the batch has the right metadata and types

    #    # expand the schema
    #    fields = []
    #    geo_names = []
    #    for f in batch.schema:
    #        if f.metadata and b"isGeometry" in f.metadata:
    #            geo_names.append(f.name)
    #            meta = f.metadata
    #            if b"index" in f.metadata:
    #                meta_copy = meta.copy()
    #                del meta_copy[b"index"]
    #                f = f.with_metadata(meta_copy)  # remove the index
    #            fields.append(f)
    #            del meta[b"isGeometry"]
    #            # TODO(oha): how to tell if there is a .z?
    #            for geometry_field in self.geometry_fields:
    #                field_name = geometry_field.format(col_name=f.name)
    #                if field_name not in batch.schema.names:
    #                    fields.append(
    #                        pa.field(geometry_field.format(col_name=f.name), pa.float64(), True, metadata=meta)
    #                    )
    #        else:
    #            fields.append(f)

    #    if not geo_names:
    #        return self.inner.insert(batch)

    #    inner_schema = pa.schema(fields)
    #    if batch.num_rows == 0:
    #        return self.inner.insert(pa.RecordBatch.from_pylist([], inner_schema))

    #    d: pd.DataFrame = batch.to_pandas()
    #    del batch  # free memory, if possible

    #    # we encode rows by rows to made it simple to create multiple columns
    #    def _encode(row):
    #        for name in geo_names:
    #            if name in row and row[name] is not None:
    #                geo = GeometryBbox.from_python_object(row[name])
    #                for geo_field, col_value in zip(self.geometry_fields, geo.get_coordinates()):
    #                    row[geo_field.format(col_name=name)] = col_value
    #                row[name] = geo.wkt
    #            else:
    #                for geo_field in self.geometry_fields:
    #                    row[str(geo_field.format(col_name=name))] = None
    #        return row

    #    for geo_name in geo_names:
    #        for geo_field in self.geometry_fields:
    #            # must add columns first, will be filled by _encode later
    #            d[geo_field.format(col_name=geo_name)] = None
    #    d = d.apply(func=_encode, axis=1)
    #    inner_batch = pa.RecordBatch.from_pandas(d, schema=inner_schema)
    #    del d

    #    return self.inner.insert(inner_batch)

    # def delete(self, query: Op, schema: pa.Schema):
    #    return self.inner.delete(query, schema)



import logging
import time
from typing import Callable, Iterator

import pandas as pd
import pyarrow as pa


class CursorException(Exception):
    """Raised when the client is required to connect again with the given cursor to fetch more data"""

    def __init__(self, cursor: str):
        self.cursor = cursor


class Cursor:
    """
    A Cursor, which reads data from a scanner and allow the user to iterate over in different flavors
    It internally uses a scanner to get the data, and perform multiple requests if a cursor is returned

    it's created by the Table.select() method
    """

    def __init__(
        self,
        scanner: Callable[[str], Iterator[pd.DataFrame]],
        schema: pa.Schema,
    ):
        self.scanner = scanner
        self.schema = schema

    def batches(self) -> Iterator[pa.RecordBatch]:
        """
        Allow the client to iterate over pyarrow Record Batches:

            for batch in tab.select().batches():
                print(batch.num_rows)
        """
        for df in self.dataframes():
            yield pa.RecordBatch.from_pandas(df, schema=self.schema)

    def dataframes(self) -> Iterator[pd.DataFrame]:
        """
        Allow the client to iterate over pandas DataFrames:

            for batch in tab.select().dataframes():
                print(len(batch))
        """
        t0 = time.perf_counter()
        cursor = ""
        ct = 0
        rows = 0
        while True:
            try:
                for df in self.scanner(cursor):
                    ct += 1
                    rows += len(df)
                    yield df
            except CursorException as e:
                cursor = e.cursor
                continue  # FIXME: Should not be raised?
            break
        logging.info("got %d batches with %d rows in %.2fs", ct, rows, time.perf_counter() - t0)

    def rows(self) -> Iterator[dict]:
        """
        Allow the client to iterate over each rows as a python dictionary:

            for row in tab.select().rows():
                print(row)
        """
        for df in self.dataframes():
            for row in df.to_dict(orient="records"):
                yield row

    def pages(self, size: int = 0) -> Iterator[list[dict]]:
        """
        Allow the client to iterate over pages of data, where each page is a list of python dictionaries.

            for page in tab.select().pages(1_000):
                print(len(page))

        If no size is specified, it will make a page for each chunk of data retrieved from the server
        """
        if size < 1:  # page based on what we get
            for df in self.dataframes():
                yield df.to_dict(orient="records")
            return

        # page based on page_size
        buf: list[dict] = []
        for df in self.dataframes():
            buf.extend(df.to_dict(orient="records"))
            while len(buf) >= size:
                yield buf[:size]
                buf = buf[size:]
        if len(buf) > 0:
            yield buf



import gzip
import io
import logging
from typing import Callable, Iterable

from odp.tabular_v2.big.big import BigCol
from odp.tabular_v2.util import Cache


class RemoteBigCol(BigCol):
    def __init__(
        self,
        uploader: Callable[[str, bytes], None],
        downloader: Callable[[str], bytes],
        root_cache: str,
    ):
        super().__init__()
        self.cache = Cache(root_cache)
        self.uploader = uploader
        self.downloader = downloader
        # TODO: make sure to not fill up the disk?

    def fetch(self, bid: str) -> bytes:
        with self.cache.key("big." + bid) as e:
            if not e.exists():
                logging.info("fetching %s", bid)
                comp = self.downloader(bid)
                e.set(comp)
            else:
                logging.info("cache hit %s", bid)
                comp = e.get()
            # if exists, use the cached version
            try:
                return gzip.decompress(comp)
            except Exception:
                e.unlink()
                raise

    def upload(self, bid: str, data: Iterable[bytes]):
        with self.cache.key("big." + bid) as e:
            buf = io.BytesIO()
            with gzip.GzipFile(fileobj=buf, mode="wb") as f:
                for d in data:
                    f.write(d)
            comp = buf.getvalue()
            self.uploader(bid, comp)
            e.set(comp)



# dto_to_stac.py
import json
import logging
import os.path
import uuid
from datetime import datetime, timezone
from typing import Any, Dict, List, Union

import pystac
import shapely

from odp.client import OdpClient
from odp.dto import DataCollectionDto, DatasetDto, ObservableDto, ResourceDto
from odp.dto.registry.observables_class_definitions import (
    static_coverage_class,
    static_geometric_coverage_class,
    static_single_value_class,
)


def dto_to_stac(
    ref: Union[uuid.UUID, str], odp_client: OdpClient
) -> Union[pystac.Catalog, pystac.Collection, pystac.Item]:
    """
    Converts a DTO object (Dataset, DataCollection, or Catalog) to a STAC object.

    Args:
        ref (UUID | str): The identifier for the DTO resource.
        odp_client (OdpClient): The ODP client instance.

    Returns:
        Union[Catalog, Collection, Item]: The corresponding STAC object.
    """
    dto = odp_client.catalog.get(ref)

    if dto.kind == DataCollectionDto.get_kind():
        dto.spec.observables = get_observables(dto, odp_client)
        return convert_collection(dto)
    elif dto.kind == DatasetDto.get_kind():
        dto.spec.observables = get_observables(dto, odp_client)
        return convert_dataset(dto)
    elif dto.kind == ObservableDto.get_kind():
        raise ValueError(f"DTO kind {dto.kind} cannot be converted to a STAC object independently")
    else:
        raise ValueError(f"Unknown DTO kind: {dto.kind}")


def get_observables(dto: ResourceDto, odp_client: OdpClient) -> List[ObservableDto]:
    """
    Get observables associated with a dataset.
    """
    try:
        observables = odp_client.catalog.list(
            {"#AND": [{"#EQUALS": ["kind", ObservableDto.get_kind()]}, {"#EQUALS": ["$spec.ref", dto.qualified_name]}]}
        )
    except Exception as e:
        logging.info(f"Error fetching observables for dataset {dto.uuid}: {e}")
        observables = []

    return list(observables)


class ObservablesInfo:
    start_time: datetime = None
    end_time: datetime = None
    geometry: Dict[str, Any] = None
    bbox: List[float] = None
    time: datetime = None

    def add_observable(self, observable: ObservableDto):
        if observable.spec.observable_class == static_coverage_class.qualified_name:
            try:
                self.start_time = datetime.fromtimestamp(observable.spec.details["value"][0])
                self.end_time = datetime.fromtimestamp(observable.spec.details["value"][1])
            except ValueError as e:
                logging.info(f"Error parsing observable {observable.uuid}: {e}")
                raise
        elif observable.spec.observable_class == static_single_value_class.qualified_name:
            try:
                value = observable.spec.details["value"]
                if isinstance(value, (int, float)):
                    self.time = datetime.fromtimestamp(value)
                else:
                    self.time = datetime.fromisoformat(value)
            except (ValueError, TypeError) as e:
                logging.info(f"Error parsing observable {observable.uuid}: {e}")
                raise ValueError(f"Error parsing observable {observable.uuid}: {e}")
        elif observable.spec.observable_class == static_geometric_coverage_class.qualified_name:
            try:
                self.geometry = observable.spec.details["value"]
                self.bbox = shapely.geometry.shape(self.geometry).bounds
            except shapely.errors.ShapelyError as e:
                logging.info(f"Error parsing observable {observable.uuid}: {e}")
                raise
        else:
            raise ValueError(f"Observable class {observable.spec.observable_class} not supported")

    def get_default_time(self) -> datetime:
        """Returns default time to current time if no time is available."""
        if self.start_time and self.end_time:
            return self.time
        elif self.time is None:
            return datetime.now(tz=timezone.utc)
        return self.time

    def get_temporal_extent(self) -> pystac.TemporalExtent:
        """Returns a pystac.TemporalExtent object based on the available time information."""
        if self.start_time and self.end_time:
            return pystac.TemporalExtent(intervals=[[self.start_time, self.end_time]])
        elif self.time:
            return pystac.TemporalExtent(intervals=[[self.time, None]])
        else:
            return pystac.TemporalExtent(intervals=[None, None])

    def get_spatial_extent(self) -> pystac.SpatialExtent:
        """Returns a pystac.SpatialExtent object based on the available spatial information.

        Defaults to the global extent if no spatial information is available.
        """
        if self.bbox:
            return pystac.SpatialExtent(bboxes=[self.bbox])
        return pystac.SpatialExtent(bboxes=[[-180, -90, 180, 90]])

    @classmethod
    def from_observables(cls, observables: List[ObservableDto]) -> "ObservablesInfo":
        info = cls()
        for observable in observables:
            try:
                info.add_observable(observable)
            except ValueError:
                continue
        return info


def convert_dataset(dto: DatasetDto, catalog_url: str = "https://app.hubocean.earth/catalog/") -> pystac.Item:
    """
    Converts a Dataset DTO to a STAC Item.
    """
    observables_info = ObservablesInfo.from_observables(dto.spec.observables)

    # Create a STAC Item
    item = pystac.Item(
        stac_extensions=[],
        href=os.path.join(catalog_url, "dataset", str(dto.metadata.name)),
        id=str(dto.uuid),
        geometry=observables_info.geometry,
        bbox=observables_info.bbox,
        properties={},
        datetime=observables_info.get_default_time(),
        start_datetime=observables_info.start_time,
        end_datetime=observables_info.end_time,
        assets={},
        collection=dto.spec.data_collection,
    )

    # If there is an associated data collection, add a link
    # TODO: This should really point to the STAC Collection definition, not the
    # collection itself or DataCollectionDto. How do we resolve this?
    if dto.spec.data_collection:
        collection_name = dto.spec.data_collection.split("/")[-1]
        collection_link = pystac.Link(
            rel=pystac.RelType.COLLECTION,
            target=os.path.join(catalog_url, "collection", collection_name),
            title=collection_name,
        )
        item.add_link(collection_link)

    # Validate STAC Item
    try:
        item.validate()
    except pystac.STACValidationError as e:
        logging.warning("Validation failed for DatasetDto %s converted to a STAC Item %s: %s", dto.uuid, item, e)
        raise

    return item


def convert_collection(
    dto: DataCollectionDto, catalog_url: str = "https://app.hubocean.earth/catalog"
) -> pystac.Collection:
    """
    Converts a DataCollection DTO to a STAC Collection.
    """
    observables_info = ObservablesInfo.from_observables(dto.spec.observables)

    extent = pystac.Extent(
        spatial=observables_info.get_spatial_extent(),
        temporal=observables_info.get_temporal_extent(),
    )

    # Determine the license name (default to "proprietary")
    license_name = "proprietary"
    license_href = None

    if dto.spec.distribution and dto.spec.distribution.license:
        license_obj = dto.spec.distribution.license  # This is an instance of License

        def normalize_license(li_name):
            li_name = li_name.replace(" ", "_")
            li_name = li_name.replace("(", "")
            li_name = li_name.replace(")", "")
            return li_name

        license_name = normalize_license(license_obj.name or license_name)  # License name
        license_href = license_obj.href  # URL to license text

    # Create pystac.Collection
    collection = pystac.Collection(
        id=str(dto.uuid),
        description=dto.metadata.description,
        href=os.path.join(catalog_url, "collection", str(dto.metadata.name)),
        title=dto.metadata.name,
        license=license_name,
        extent=extent,
    )

    # Add license link if href is available
    if license_href:
        license_link = pystac.Link(
            rel="license",
            target=license_href,
            title=f"License: {license_name}",
        )
        collection.add_link(license_link)

    # Validate STAC collection
    try:
        collection.validate()
    except pystac.STACValidationError as e:
        logging.warning(
            "Validation failed for DatasetDto %s converted to a STAC Collection %s: %s", dto.uuid, collection, e
        )
        raise

    return collection


# TESTING
if __name__ == "__main__":
    # TEST #1

    # UUID of a public dataset that we're using in testing:
    # https://app.hubocean.earth/catalog/dataset/wwf-oceansfutures-species
    dataset_uuid = "00acf0c5-d0ff-4f5e-a75c-ea691b9cc61b"

    client = OdpClient()
    dataset = client.catalog.get(dataset_uuid)

    print("\n--- Printing Test Dataset ---")
    # Expected:
    # {
    #     "spec": {
    #         "storage_class": "registry.hubocean.io/storageClass/tabular",
    #         "storage_controller": "registry.hubocean.io/storageController/storage-tabular",
    #         "data_collection": "catalog.hubocean.io/dataCollection/wwf-oceansfutures",
    #         "maintainer": {
    #             "contact": "Sarah Glaser <sarah.glaser@wwfus.org>",
    #             "organisation": "World Wildlife Fund US",
    #         },
    #         "citation": null,
    #         "documentation": ["https://www.oceansfutures.org/methodology"],
    #         "attributes": [],
    #         "facets": null,
    #         "tags": [
    #             "marine resources",
    #             "fish species",
    #             "Sea Around Us",
    #             "species distribution",
    #             "fisheries management",
    #             "species composition",
    #             "catch data",
    #             "commercial fishing",
    #             "EEZ",
    #             "SAU",
    #         ],
    #     },
    #     "kind": "catalog.hubocean.io/dataset",
    #     "version": "v1alpha3",
    #     "metadata": {
    #         "name": "wwf-oceansfutures-species",
    #         "display_name": "WWF Oceans Futures - Key Commercial Fish Species by EEZ",
    #         "description": "WWF Oceans Futures Phase 1 dataset detailing the top commercial fish species for each
    #         Exclusive Economic Zone (EEZ). Includes common species names and their relative importance as a percentage
    #         of the top 5 species caught in each EEZ, using Sea Around Us (SAU) EEZ identifiers.",
    #         "uuid": "00acf0c5-d0ff-4f5e-a75c-ea691b9cc61b",
    #         "labels": {
    #             "huboceanData": true,
    #             "raw_file_names": ["fish_species.csv"],
    #             "catalog.hubocean.io/released": true,
    #         },
    #         "owner": "16f81607-9bc0-455c-a77f-618ea68fa9a0",
    #     },
    #     "status": {
    #         "num_updates": 0,
    #         "created_time": "2024-12-20T12:12:11.653220",
    #         "created_by": "16f81607-9bc0-455c-a77f-618ea68fa9a0",
    #         "updated_time": "2024-12-20T12:12:11.653220",
    #         "updated_by": "16f81607-9bc0-455c-a77f-618ea68fa9a0",
    #         "deleted_time": null,
    #         "deleted_by": null,
    #     },
    # }
    print(dataset.json())

    # Convert DatasetDto to STAC Item
    print(f"Converting DatasetDto {dataset_uuid} to a STAC Item")
    stac_item = dto_to_stac(dataset_uuid, client)
    print("Validating STAC Item")
    stac_item.validate()

    # Print STAC Item
    print("\n--- Printing STAC Item Spec ---")
    print(json.dumps(stac_item.to_dict(), indent=4))

    # TEST #2

    # UUID of a public data collection that we're using in testing:
    # https://app.hubocean.earth/catalog/collection/moura2016-amazoncoral
    data_collection_uuid = "07c35450-3c95-4014-b65a-953269f69b77"

    data_collection = client.catalog.get(data_collection_uuid)

    print("\n--- Printing Test Data Collection ---")
    # Expected:
    # {
    #     "spec": {
    #         "distribution": {
    #             "published_by": {
    #                 "contact": "Thompson, Fabiano L. <fabianothompson1@gmail.com>",
    #                 "organisation": "Instituto de Biologia, Universidade Federal do Rio de Janeiro (UFRJ)",
    #             },
    #             "published_date": "2016-04-22T00:00:00",
    #             "website": "https://www.science.org/doi/full/10.1126/sciadv.1501252",
    #             "license": {
    #                 "name": "CC-BY-4.0",
    #                 "href": "https://creativecommons.org/licenses/by-nc/4.0/",
    #                 "full_text": null,
    #             },
    #         },
    #         "tags": [
    #             "Amazon River",
    #             "marine biology",
    #             "benthic habitats",
    #             "South Atlantic",
    #             "oceanography",
    #             "rhodoliths",
    #             "continental shelf",
    #             "mesophotic reefs",
    #             "reef system",
    #             "fisheries",
    #             "carbonate structures",
    #             "French Guiana",
    #             "ecosystem services",
    #             "climate change",
    #             "biogenic reefs",
    #             "Brazil",
    #             "sponge communities",
    #             "marine biodiversity",
    #             "river plume",
    #             "marine ecosystems",
    #         ],
    #         "facets": null,
    #     },
    #     "kind": "catalog.hubocean.io/dataCollection",
    #     "version": "v1alpha1",
    #     "metadata": {
    #         "name": "moura2016-amazoncoral",
    #         "display_name": "Amazon River Mouth Reef System (Moura et al. 2016)",
    #         "description": "Comprehensive dataset collection revealing an extensive and previously unknown reef system
    #         at the Amazon River mouth. The collection includes data from multidisciplinary surveys conducted between
    #         2010-2014 that discovered a significant carbonate system spanning approximately 9,500 km\u00b2 between
    #         French Guiana and Brazil. The datasets document reef structures, rhodolith beds, sponge communities, and
    #         fishing activities, along with oceanographic measurements and the seasonal influence of the Amazon River
    #         plume. This collection provides crucial baseline data about reef ecosystems functioning under marginal
    #         conditions, with implications for understanding reef systems' responses to climate change.",
    #         "uuid": "07c35450-3c95-4014-b65a-953269f69b77",
    #         "labels": {
    #             "etl_config": {
    #                 "type": "BlobStorage",
    #                 "azure_blob_container_name": "moura2016-amazoncoral",
    #                 "azure_blob_connection_string_secret_name": "sadatapipelines-connection-string",
    #             },
    #             "huboceanData": true,
    #             "catalog.hubocean.io/released": true,
    #         },
    #         "owner": "16f81607-9bc0-455c-a77f-618ea68fa9a0",
    #     },
    #     "status": {
    #         "num_updates": 1,
    #         "created_time": "2025-01-24T14:19:10.393967",
    #         "created_by": "16f81607-9bc0-455c-a77f-618ea68fa9a0",
    #         "updated_time": "2025-01-29T08:25:30.792966",
    #         "updated_by": "16f81607-9bc0-455c-a77f-618ea68fa9a0",
    #         "deleted_time": null,
    #         "deleted_by": null,
    #     },
    # }
    print(data_collection.json())

    # Convert DataCollectionDto to STAC Collection
    print(f"Converting DataCollectionDto {data_collection_uuid} to a STAC Collection")
    stac_collection = dto_to_stac(data_collection_uuid, odp_client=client)
    print("Validating STAC Collection")
    stac_collection.validate()

    # Print STAC Collection
    print("\n--- Printing STAC Collection Spec ---")
    print(json.dumps(stac_collection.to_dict(), indent=4))



from datetime import datetime
from typing import Any, List, Optional, Set

from ..common.contact_info import ContactInfo
from ..common.license import License
from ..resource import ResourceDto, ResourceSpec
from ..resource_registry import DEFAULT_RESOURCE_TYPE_REGISTRY
from .observable import ObservableDto
from odp.util.cheapdantic import BaseModel, Field


class Distribution(BaseModel):
    """Distribution information"""

    published_by: ContactInfo
    """Publisher information"""

    published_date: datetime
    """Date of first published"""

    website: str
    """Distribution website"""

    license: Optional[License] = None
    """Dataset license information"""


class DataCollectionSpec(ResourceSpec):
    distribution: Optional[Distribution] = None
    """Information on how the dataset was distributed"""

    tags: Set[str] = Field(default_factory=set)
    """Tags for the dataset"""

    facets: Optional[dict[str, Any]] = None
    """Facets for the dataset"""

    observables: List[ObservableDto] = Field(default_factory=list)
    """Observables for the data collection"""


class DataCollectionDto(ResourceDto):
    _kind: str = "catalog.hubocean.io/dataCollection"
    _version: str = "v1alpha1"

    spec: DataCollectionSpec


DEFAULT_RESOURCE_TYPE_REGISTRY.add(DataCollectionDto)
